from turbogears.validators import Schema, Invalid, FormValidator, \
    String
from tgcaptcha2 import controller
import gettext
from turbogears import config
from datetime import datetime

_ = gettext.gettext


class ValidCaptchaInput(FormValidator):

    _nonce_checker = False

    def __init__(self, checker, *args, **kwargs):
        super(FormValidator, self).__init__(*args, **kwargs)
        self._use_nonce = checker

    messages = {'incorrect': _("Incorrect value."),
                'notfresh': _('Already used.'),
                'timeout': _("Too much time elapsed. Please try again.")}

    __unpackargs__ = ('captchahidden', 'captchainput')

    timeout = int(config.get('tgcaptcha2.timeout', 5))

    def validate_python(self, field_dict, state):
        ctrl = controller.CaptchaController()

        hidden = str(field_dict['captchahidden'])
        input_val = str(field_dict['captchainput'])
        try:
            payload = ctrl.model_from_payload(hidden)
        except:
            raise Invalid(self.message('incorrect', state), field_dict, state)
        if payload.plaintext != input_val:
            raise Invalid(self.message('incorrect', state), field_dict, state)
        elapsed = datetime.utcnow() - payload.created
        if elapsed.seconds > self.timeout * 60:
            raise Invalid(self.message('timeout', state), field_dict, state)
        if not self._use_nonce(payload.nonce):
            raise Invalid(self.message('notfresh', state), field_dict, state)


class CaptchaFieldValidator(Schema):

    captchahidden = String(not_empty=True)
    captchainput = String(not_empty=True)

    _used_nonces = []
    def _basic_nonce_checker(self, nonce):
        if nonce in self._used_nonces:
            return False
        self._used_nonces.append(nonce)
        return True

    def __init__(self, nonce_checker=None, *args, **kwargs):
        super(Schema, self).__init__(*args, **kwargs)
        if nonce_checker:
            checker = nonce_checker
        else:
            checker = self._basic_nonce_checker
        self.chained_validators = [ValidCaptchaInput(checker, 'captchahidden', 'captchainput')]
